*/
AUTOR: ANDRE LUIS MALDONADO DANIEL
EMAIL: andredaniel@alunos.utfpr.edu.br
DISCIPLINA: SISTEMAS EMBARCADOS
PROJETO: CONTROLE DE PLACA RELE
*/

#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <bcm2835.h>
#include <mysql/mysql.h>
#include <stdlib.h>

// Variaveis globais
// Handler MYSQL
MYSQL* CONEXAO;

// Dados do usuario
char USUARIO[] = "root";
char SENHA[] = "mysqllt38c";

// Dados do banco de dado 
char HOST[] = "localhost";
char DATABASE[] = "lt38c";
char TB_ESCRITA[] = "pinwr";
char TB_LEITURA[] = "pinrw";
unsigned int PORT = 0;
char* UNIX_SOCKET = NULL;
unsigned long CLIENT_FLAG = 0;

// Tabelas
// Mapeamento dos pinos fisicos para BCM
const unsigned char rpi_para_gpio[41] = 
{	
 0,  0,  0,  2, 0,  3,  0,  4, 14,  0, 
15, 17, 18, 27, 0, 22, 23,  0, 24, 10,
 0,	 9, 25, 11, 8,  0,  7,  0,  0,  5, 
 0,  6, 12, 13, 0, 19, 16, 26, 20,  0,
21
};

// Pinos que são saída
const unsigned char pinos_gpio_saida[4] =
{
	4,6,22,26
};

// Pinos que são entrada
const unsigned char pinos_gpio_entrada[2] =
{
	20,21
};

void logs(const char* mensagem);
bool conectar();
void desconectar();
bool ligar_bcm();
bool desligar_bcm();
void inicializar_pinos();
bool mudar_estado_pino(unsigned char comando, unsigned char pino_rpi);
void pino_HIGH(unsigned char pino_rpi);
void pino_LOW(unsigned char pino_rpi);
void adicionar_comando_db(unsigned char comando, unsigned char pino, const char* email);
bool requisicao_db(const char* requisicao);
void executar_comandos_db();
void leitura_pinos_para_db();

int main(int argc, char** argv)
{
	logs("Iniciando programa\n");
	char buffer[512];
	
	if(argc == 1)
	{
		printf("Numero de argumentos invalido\n");
		printf("Uso do programa: -[comando] [modo] [pino] [email]\n");
		printf("Para adicionar um comando novo use o comando ac\n");
		printf("Para iniciar a leitura e execução dos comandos do banco de dados use o comando s\n");
		logs("Numero de argumentos invalido\n");
		return 1;
	}
	
	if(strcmp(argv[1], "-ac") == 0)
	{
		logs("Iniciando execução '-ac'\n");
		CONEXAO = mysql_init(NULL);
		
		if(argc < 4)
		{
			printf("Numero de argumentos invalido\n");
			logs("Numero de argumentos invalido\n");
			return 1;
		}
		if(conectar())
		{
			unsigned char comando = atoi(argv[2]);
			unsigned char pino_rpi = atoi(argv[3]);
			
			if(comando > 2 || comando < 0)
			{
				sprintf(buffer,"Comando [%u] invalido\n", comando);
				printf(buffer);
				logs(buffer);
				desconectar();
				return 1;
			}

			
			if (rpi_para_gpio[pino_rpi] == pinos_gpio_saida[0] ||
			rpi_para_gpio[pino_rpi] == pinos_gpio_saida[1] ||
			rpi_para_gpio[pino_rpi] == pinos_gpio_saida[2] ||
			rpi_para_gpio[pino_rpi] == pinos_gpio_saida[3])
			{
				adicionar_comando_db(comando, pino_rpi, argv[4]);
				desconectar();
				return 0;
			}
			
			sprintf(buffer,"Pino RPI:[%u], GPIO[%u] invalido\n", pino_rpi, rpi_para_gpio[pino_rpi]);
			printf(buffer);
			logs(buffer);
			desconectar();
			return 1;
		}
		else
		{
			return 1;
		}
		
	}
	if(strcmp(argv[1], "-s") == 0)
	{
		printf("Iniciando execução\nAperte CTRL+C a qualquer momento para sair\n");
		logs("Iniciando execução '-s'\n");
		while(1)
		{
			CONEXAO = mysql_init(NULL);
			if(conectar())
			{
				if(ligar_bcm())
				{
					inicializar_pinos();
					executar_comandos_db();
					leitura_pinos_para_db();
					if(!desligar_bcm())
					{
						return 1;
					}
				}
				desconectar();
			}
			
			else 
			{
				return 1;
			}
			delay(50);
		}
	}
	sprintf(buffer, "Comando '%s' incorreto\n",argv[1]);
	printf(buffer);
	logs(buffer);
	return 1;
}

// Apende ao arquivo uma mensagem. Caso o arquivo não exista ele é criado.
void logs (const char* mensagem)
{
	FILE* log = fopen("/tmp/projeto_log.txt","a");
	if(log == NULL)
	{
		printf("Erro ao abrir arquivo de LOG\n");
		return;
	}
	fprintf(log,"%s",mensagem);
	fclose(log);
}


// Conecta ao banco de dados e retorna 1 caso consiga.
bool conectar()
{
	char buffer[512];
	
	if(!mysql_real_connect(CONEXAO, HOST, USUARIO, SENHA, DATABASE, PORT, UNIX_SOCKET,CLIENT_FLAG))
	{
		logs("Falha de conexao com o banco de dados\n");
		sprintf(buffer,"Erro %d : %s\n", mysql_errno(CONEXAO), mysql_error(CONEXAO));
        printf(buffer);
		logs(buffer);
		return 0;
	}
	else
	{
		logs("Conectado ao banco de dados.\n");
		return 1;
	}
}

// Desconeta do banco de dados, não tem retorno.
void desconectar()
{
	mysql_close(CONEXAO);
	logs("Desconectado do banco de dados\n");
}

// Inicia a biblioteca BCM2835, retorna 1 em caso positivo.
bool ligar_bcm()
{
	if(!bcm2835_init())
	{
		printf("Erro: bcm2835 não conseguiu inicializar\n");
		logs("Erro: bcm2835 não conseguiu inicializar\n");
		return 0;
	}
	else
	{
		logs("bcm2835 iniciado\n");
		return 1;
	}
}

// Desativa a conexão BCM2835, retorna 1 em caso positivo
bool desligar_bcm()
{
	if(!bcm2835_close())
	{
		printf("Erro: bcm2835 não conseguiu desligar\n");
		logs("Erro: bcm2835 não conseguiu desligar\n");
		return 0;
	}
	else
	{
		logs("bcm2835 desligado\n");
		return 1;
	}
}

// Inicializa os pinos de entrada e saída.
void inicializar_pinos()
{
	logs("Inicializando pinos\n");
	char buffer[512];
	logs("Pinos de saida:\n");
	
	for(unsigned char i = 0; i < 4; i++)
	{
		sprintf(buffer, "Inicializando pino %d como saída\n",pinos_gpio_saida[i]);
		//printf(buffer);
		logs(buffer);
		bcm2835_gpio_fsel(pinos_gpio_saida[i], BCM2835_GPIO_FSEL_OUTP);
	}
	
	logs("Pinos de entrada:\n");
	for(unsigned char i = 0; i < 2; i++)
	{
		sprintf(buffer, "Inicializando pino %d como entrada\n",pinos_gpio_entrada[i]);
		//printf(buffer);
		logs(buffer);
		bcm2835_gpio_fsel(pinos_gpio_entrada[i], BCM2835_GPIO_FSEL_INPT);
	}
	logs("Inicializacao de pinos concluída\n");
}

// Muda o estado do pino de acordo com o comando recebido, retorna 0 em caso de comando invalido.
bool mudar_estado_pino(unsigned char comando, unsigned char pino_rpi)
{
	logs("Mudando estado dos pinos:\n");
	char buffer[512];
	switch(comando)
	{
		case 1:
		sprintf(buffer, "Mudando o estado do pino %d para HIGH\n", rpi_para_gpio[pino_rpi]);
		logs(buffer);
		pino_HIGH(pino_rpi);
		return 1;
		
		case 2:
		sprintf(buffer, "Mudando o estado do pino %d para LOW\n", rpi_para_gpio[pino_rpi]);
		logs(buffer);
		pino_LOW(pino_rpi);
		return 1;
		
		default:
		sprintf(buffer, "Comando invalido\n", rpi_para_gpio[pino_rpi]);
		logs(buffer);
		return 0;
	}
}

// Muda o estado do pino para LOW(0).
void pino_LOW(unsigned char pino_rpi)
{
    bcm2835_gpio_write(rpi_para_gpio[pino_rpi], LOW);
}

// Muda o estado do pino para HIGH(1).
void pino_HIGH(unsigned char pino_rpi)
{
    bcm2835_gpio_write(rpi_para_gpio[pino_rpi], HIGH);
}

// Faz as requisições(query) para o banco de dados, retorna 0 em caso de erro.
bool requisicao_db(const char* requisicao)
{
	logs("Fazendo requsição ao banco de dados: ");
	logs(requisicao);
	logs("\n");
	
	if(mysql_query(CONEXAO, requisicao))
	{
		char buffer[512];
		sprintf(buffer,	"Erro: %s [%d].\n", mysql_error(CONEXAO), mysql_errno(CONEXAO));
		logs("Requisição encontrou um erro: ");
		logs(buffer);
		return 0;
	}
	else
	{
		return 1;
	}
}


// Adiciona comandos para serem executados no banco de dados.
void adicionar_comando_db(unsigned char comando, unsigned char pino, const char* email)
{
	printf("Adicionando comando no banco de dados\n");
	logs("Adicionando comando no banco de dados\n");
	char buffer[512];
	sprintf(buffer, "INSERT INTO pinwr(comando,pino,sync,tempo,datah,email) VALUES (%u,%u,1,12,current_timestamp(),'%s')",comando,pino,email);
	
	if(requisicao_db(buffer))
	{
		printf("Comando inserido com sucesso\n");
		logs("Comando inserido com sucesso\n");
		return;
	}
	else
	{
		return;
	}
}

// Busca comandos ainda não executados no banco de dados e os executa. Atualiza estado sync no ao faze-lo.
void executar_comandos_db()
{
	logs("Executando comandos do banco de dados\n");
	MYSQL_RES* resultado = NULL;
	MYSQL_ROW linha;
	char buffer[512];
	
	if(requisicao_db("SELECT * FROM `pinwr` WHERE sync = 1;"))
	{
		resultado = mysql_store_result(CONEXAO);
	}
	else
	{
		return;
	}
	
	if(resultado == NULL)
	{
		logs("Erro: requisição não devolveu resultados\n");
		logs("Liberando memoria do resultado da busca\n");
		mysql_free_result(resultado);
		return;
	}
	
	while(linha = mysql_fetch_row(resultado))
	{
		unsigned long id = atoi(linha[0]);
		unsigned char comando = atoi(linha[1]);
		unsigned int pino_rpi = atoi(linha[2]);
		//printf("RPI: %d GPIO: %d\n",pino_rpi,rpi_para_gpio[pino_rpi]);
		if(comando > 2 || comando < 0)
		{
			sprintf(buffer,"Comando [%u] do banco de dado e invalido\n", comando);
			logs(buffer);
		}
		if (rpi_para_gpio[pino_rpi] == pinos_gpio_saida[0] ||
			rpi_para_gpio[pino_rpi] == pinos_gpio_saida[1] ||
			rpi_para_gpio[pino_rpi] == pinos_gpio_saida[2] ||
			rpi_para_gpio[pino_rpi] == pinos_gpio_saida[3])
		{
			if(mudar_estado_pino(comando,pino_rpi))
			{
				sprintf(buffer, "UPDATE %s.%s SET sync = 0 WHERE id = %lu;",DATABASE,TB_ESCRITA,id);
				if(requisicao_db(buffer))
				{
					sprintf(buffer,"Tabela %s do banco de dados %s executou comando e atualizou com sucesso\n",TB_ESCRITA,DATABASE);
					logs(buffer);
				}
				else
				{
					sprintf(buffer,"Erro ao atualizar tabela %s do banco de dados %s\n",TB_ESCRITA,DATABASE);
					logs(buffer);
				}
			}
		}
		else
		{
			sprintf(buffer,"Pino RPI:[%u], GPIO[%u] do banco de dados e invalido\n", pino_rpi, rpi_para_gpio[pino_rpi]);
			logs(buffer);
		}
	}
	logs("Todos os comandos foram executados\n");
	logs("Liberando memoria do resultado da busca\n");
	mysql_free_result(resultado);
}

// Lê os pinos selecionados como entradas e atualiza o banco dados com os valores lidos.
void leitura_pinos_para_db()
{
	logs("Lendo pinos e enviando para banco de dados\n");
	MYSQL_RES* resultado = NULL;
	MYSQL_ROW linha;
	char buffer[512];
	if(requisicao_db("SELECT * FROM `pinrw`;"))
	{
		resultado = mysql_store_result(CONEXAO);
	}
	else
	{
		return;
	}
	
	if(resultado == NULL)
	{
		logs("Erro: requisição não devolveu resultados\n");
		logs("Liberando memoria do resultado da busca\n");
		mysql_free_result(resultado);
		return;
	}
	while(linha = mysql_fetch_row(resultado))
	{
		unsigned long id = atoi(linha[0]);
		unsigned char pino_rpi = atoi(linha[1]);
		unsigned char leitura = bcm2835_gpio_lev(rpi_para_gpio[pino_rpi]);
		sprintf(buffer,"UPDATE %s.%s SET valor = %lu WHERE id = %lu;", DATABASE, TB_LEITURA, leitura, id);
		if(requisicao_db(buffer))
		{
			sprintf(buffer,"Tabela %s do banco de dados %s executou comando e atualizou com sucesso\n",TB_LEITURA,DATABASE);
			logs(buffer);
		}
		else
		{
			sprintf(buffer,"Erro ao atualizar tabela %s do banco de dados %s\n",TB_LEITURA,DATABASE);
			logs(buffer);
		}
	}
	logs("Leitura concluída\n");
	logs("Liberando memoria do resultado da busca\n");
	mysql_free_result(resultado);
}